import extend from 'extend'
import loremIpsum from 'lorem-ipsum'
import clc from 'cli-color'
import {parse} from 'node-sqlparser'
import fse from 'fs-extra'

import {
  config,
  cli,
  log,
  folderUtils,
  fileUtils,
  FileParser,
  fileAttr,
  getAttr
} from '../'

export default class Utils {

  constructor() {
    this._form = {
      default: {
        item:[]
      }
    }
    this._key = []
  }

  /**
   * Get all input from a template
   * @return {Array} array of input form
   */
  get form(){
    return this._form
  }

  /**
   * Check if key is not is the form array
   * @param  {[type]} key [description]
   * @return {[type]}     [description]
   */
  dontHaveKey(key){
    return typeof this._key[key] === "undefined" || this._key[key] === null
  }

  /**
   * Add entry to abe engine form
   * @param {String} type        textarea | text | meta | link | image | ...
   * @param {String} key         unique ID, no space allowed
   * @param {String} desc        input description
   * @param {Int}    maxLength   maximum characteres allowed inside input
   * @param {String} tab         tab name
   * @param {String} jsonValue   
   * @return {Void}
   */
  add(type, key, desc, maxLength, tab, jsonValue, source, display) {
    var value = ''
    var block = ''
    if(typeof jsonValue !== 'undefined' && jsonValue !== null) {
      value = jsonValue
    }

    if(key.indexOf('[') < 0 && key.indexOf('.') > -1) {
      block = key.split('.')[0]
    }

    if(tab === '' || typeof tab === 'undefined' || tab === null) tab = 'default'
    else if(typeof this._form[tab] === 'undefined' || this._form[tab] === null) this._form[tab] = {item:[]}
    this._key[key] = true // save key for dontHaveKey()
    this._form[tab].item.push({
      type: type,
      key: key.replace(/\./, '-'),
      description: desc,
      maxLength: maxLength,
      placeholder: '',
      value: value,
      tab: tab,
      source: source,
      display: display,
      block: block
    })
  }

  isBlock(str){
    return str.indexOf('[') < 0 && str.indexOf('.') > 0
  }

  /**
   * Test if a string don't contains string key from ABE block statement
   * @param  {String}  str string to test
   * @return {Boolean} true = this is not a block content
   */
  isSingleAbe(str, text){
    return  !new RegExp('#each(.)+?' + getAttr(str, 'key').split('.')[0]).test(text) &&
            str.indexOf('{{#') < 0 &&
            str.indexOf('#each') < 0 &&
            str.indexOf('{{/') < 0 &&
            str.indexOf('/each') < 0 &&
            str.indexOf('attrAbe') < 0
  }

  /**
   * Test if a string contains string key from ABE block statement
   * @param  {String}  str string to test
   * @return {Boolean} true = this is a block content
   */
  isBlockAbe(str) {
    return str.indexOf('abe') > -1 && getAttr(str, 'key').indexOf('.') > -1
  }

  /**
   * Test if a string contains string key from {{#each}} block statement
   * @param  {String}  str string to test
   * @return {Boolean} true = this is a block content
   */
  isEachStatement(str) {
    return str.indexOf('#each') > -1 || str.indexOf('/each') > -1
  }

  /**
   * Encode / Escape && add data-abe attributs
   * @param  {String} block
   * @return {String} escaped string
   */
  encodeAbe(block){
    var matchAbe = block.match(/>\s*\{\{abe .*\}\}/g)
    if(matchAbe){
      for (var i = 0; i < matchAbe.length; i++)
        block = block.replace(
          matchAbe[i],
          ' data-abe="'  + getAttr(matchAbe[i], 'key').replace('.', '[0]-') + '" >'
        )
    }
    matchAbe = block.match(/( [A-Za-z0-9\-\_]+="*{{.*}})/g)
    if(matchAbe){
      for (var i = 0; i < matchAbe.length; i++)
        if(typeof matchAbe !== 'undefined' && matchAbe !== null){
          block = block.replace(
            matchAbe[i],
            ' data-abe-attr="'  + (matchAbe[i].split('=')[0]).trim() + '"' +
            ' data-abe="' + getAttr(matchAbe[i], 'key').replace('.', '[0]-') + '" ' + matchAbe[i]
          )
        }
    }
    return escape(block)
  }

  /**
   * Add some stuff like style / script before closing </body> tag
   * @param  {String} text html page
   * @return {String} text + some sugar stuff added on the fly
   */
  insertDebugtoolUtilities(text){
    return text.replace(
      /<\/body>/,
        `<style>
          body [data-abe]{ transition: box-shadow 600ms ease-in-out; box-shadow: 0; }
          body .select-border{ border-color: #007CDE; box-shadow: 0 3px 13px #7CBAEF; }
          body img.display-attr:before { content: attr(alt); }
          body a.display-attr:before { content: attr(title); }
          body .display-attr:before { position: absolute; display: block; z-index: 555; font-size: 10px; background-color: rgba(255, 255, 255, 0.75); padding: 2px 5px; color: #5D5D5D; }
          .hidden-abe{ display: none!important; }
        </style>
      </body>`
    )
  }

  lorem(type, v) {
    var lorem = ''
    if(type === 'text') {
      lorem = loremIpsum({
        units: 'sentences' // Generate words, sentences, or paragraphs.
        , sentenceLowerBound: 5
        , sentenceUpperBound: 10
      })
    }else if(type === 'link') {
      lorem = 'http://www.google.com'
    }else if(type === 'image' || type === 'file') {
      var width = getAttr(v, 'width')
      width = (width !== '') ? width : 300
      var height = getAttr(v, 'height')
      height = (height !== '') ? height : 300
      lorem = `http://placehold.it/${height}x${width}`
    }else if(type === 'textarea' || type === 'rich') {
      lorem = loremIpsum({
        units: 'paragraphs' // Generate words, sentences, or paragraphs.
        , paragraphLowerBound: 3
        , paragraphUpperBound: 7
      })
    }
    return lorem
  }

  static addMetas(tpl, json, type, obj = {}, user = {}) {
    let meta = config.meta.name

    json[meta] = extend({}, json[meta])
    var currentDate = new Date()
    var username = user.name || ''

    if(typeof json[meta].date === 'undefined' || json[meta].date === null) {
      json[meta].date = currentDate
      json[meta].author = username
    }
    json[meta].latest = {
      date: currentDate,
      author: username
    }

    if(typeof json[meta][type] === 'undefined' || json[meta][type] === null) {
      json[meta][type] = JSON.parse(JSON.stringify(obj))
      json[meta][type].date = currentDate
      json[meta][type].author = username
    }
    json[meta][type].latest = JSON.parse(JSON.stringify(obj))
    json[meta][type].latest.date = currentDate
    json[meta][type].latest.author = username
  }

  static recurseWhere(where, operator = '') {
    var arr = []
    var obj = {}
    var operatorLeft = operator
    var operatorRight = operator

    if(typeof where.left !== 'undefined' && where.left !== null
      && typeof where.right !== 'undefined' && where.right !== null
      && typeof where.operator !== 'undefined' && where.operator !== null) {
      // SQL WHERE
      
      if(typeof where.left.column !== 'undefined' && where.left.column !== null
        && typeof where.right.column !== 'undefined' && where.right.column !== null
        && typeof where.operator !== 'undefined' && where.operator !== null) {
        obj.left = where.left.column
        obj.right = where.right.column
        obj.compare = where.operator
        obj.operator = operator
        arr.push(obj)
      }else {
        if(typeof where.left.left.type !== 'undefined' && where.left.left.type !== null) {
          operator = where.operator
        }
      }

      if(typeof where.left.left !== 'undefined' && where.left.left !== null) {
        arr = arr.concat(Utils.recurseWhere(where.left, operator))
      }

      if(typeof where.right.right !== 'undefined' && where.right.right !== null) {
        arr = arr.concat(Utils.recurseWhere(where.right, operator))
      }
    }

    return arr
  }

  static handleSqlRequest(str) {
    var from = /from ([A-Za-z0-9\-\_\/, ]+)/.exec(str)

    if(typeof from !== 'undefined' && from !== null && typeof from[1] !== 'undefined' && from[1] !== null) {
      var escapedFrom = from[1].replace(/\//g, '___abe___')
      str = str.replace(from[1], escapedFrom)
    }

    var request = parse(str)
    var reconstructSql = ''

    // SQL TYPE
    var type = ''
    if(typeof request.type !== 'undefined' && request.type !== null) {
      type = request.type
    }
    reconstructSql += `${type} `

    // SQL COLUMNS
    var columns = []
    if(typeof request.columns !== 'undefined' && request.columns !== null) {
      if(request.columns === '*') {
        columns.push('*')
      }else {
        Array.prototype.forEach.call(request.columns, (item) => {
          columns.push(item.expr.column)
        })
      }
    }
    reconstructSql += `${JSON.stringify(columns)} `

    // SQL FROM
    var from = []
    if(typeof request.from !== 'undefined' && request.from !== null) {
      Array.prototype.forEach.call(request.from, (item) => {
        from.push(item.table)
      })
    }else {
      from.push('*')
    }
    reconstructSql += `from ${JSON.stringify(from)} `


    var where
    if(typeof request.where !== 'undefined' && request.where !== null) {
      where = Utils.recurseWhere(request.where)
      reconstructSql += 'where '
      Array.prototype.forEach.call(where, (w) => {
        reconstructSql += `${w.operator} ${w.left} ${w.compare} ${w.right} `
      })
    }

    var limit = -1
    if(typeof request.limit !== 'undefined' && request.limit !== null) {
      limit = request.limit[request.limit.length - 1].value
    }

    var orderby
    if(typeof request.orderby !== 'undefined' && request.orderby !== null && request.orderby.length > 0) {
      orderby = {
        column: request.orderby[0].expr.column,
        type: request.orderby[0].type
      }
      reconstructSql += `ORDER BY ${orderby.column} ${orderby.type} `
    }

    log.write('abe-utils-handleSqlRequest', reconstructSql)

    return {
      type: type,
      columns: columns,
      from: from,
      where: where,
      string: reconstructSql,
      limit: limit,
      orderby: orderby
    }
  }

  static sortByDateDesc(a, b) {
    var jsonA = fse.readJsonSync(a.path)
    var jsonB = fse.readJsonSync(b.path)
    let metaA = config.meta.name
    let metaB = config.meta.name

    var dateA = (jsonA[metaA]) ? new Date(jsonA[metaA].latest.date) : new Date()
    var dateB = (jsonB[metaB]) ? new Date(jsonB[metaB].latest.date) : new Date()
    if(dateA < dateB) {
      return 1
    }else if(dateA > dateB) {
      return -1
    }
    return 0
  }

  static sortByDateAsc(a, b) {
    var jsonA = fse.readJsonSync(a.path)
    var jsonB = fse.readJsonSync(b.path)
    let metaA = config.meta.name
    let metaB = config.meta.name
    var dateA = new Date(jsonA[metaA].latest.date)
    var dateB = new Date(jsonB[metaB].latest.date)
    if(dateA > dateB) {
      return 1
    }else if(dateA < dateB) {
      return -1
    }
    return 0
  }

  static getDataSource(str) {
    var res = str.substring(str.indexOf('value=') + 7, str.length)

    var reg = /([^'"]*=[\s\S]*?}})/g
    var matches = res.match(reg)
    if(typeof matches !== 'undefined' && matches !== null) {
      Array.prototype.forEach.call(matches, (matche) => {
        res = res.replace(matche, '')
      })
    }else {
      res = res.replace('}}', '')
    }

    return res.substring(0, res.length-1)
  }

  static getDataRequest(website, match) {
    var request = Utils.handleSqlRequest(getAttr(match, 'request')),
        limit = 0,
        res = []

    let data = config.data.url
    var files = FileParser.getFiles(fileUtils.concatPath(config.root, data), true, 99, /\.json/)

    if(typeof request.orderby !== 'undefined' && request.orderby !== null) {
      if(request.orderby.column.toLowerCase() === 'date') {
        if(request.orderby.type === 'ASC') {
          files.sort(Utils.sortByDateAsc)
        }else if(request.orderby.type === 'DESC') {
          files.sort(Utils.sortByDateDesc)
        }
      }
    }

    Array.prototype.forEach.call(files, (file) => {
      var shouldAdd = false
      var folderPath = folderUtils.getFolderPath(file.path)
      var json

      // remove not published file
      var attr = fileAttr.get(file.path)
      if(typeof attr.d !== 'undefined' && attr.d !== null) {
        return;
      }

      // check from
      if(typeof request.from !== 'undefined' && request.from !== null) {
        Array.prototype.forEach.call(request.from, (from) => {
          from = from.replace(/___abe___/g, '/')
          if(from === '*' || from === '/') {
            shouldAdd = true
          }else if(folderPath.indexOf(from) > -1) {
            shouldAdd = true
          }
        })
      }else {
        shouldAdd = true
      }

      // check from
      if(typeof request.where !== 'undefined' && request.where !== null) {
        shouldAdd = false
        json = fse.readJsonSync(file.path)
        let meta = config.meta.name

        var template = FileParser.getTemplate(json[meta].template)
        Array.prototype.forEach.call(request.where, (where) => {
          if(where.left === 'template') {
            switch(where.compare) {
              case '=':
                if(where.right === template) {
                  shouldAdd = true
                }
                break;
              case '!=':
                if(where.right === template) {
                  shouldAdd = false
                }
                break;
              case 'LIKE':
                if(template.indexOf(where.right) > -1) {
                  shouldAdd = true
                }
                break;
              default:
                break;
            }
          }
        })
      }

      if(shouldAdd) {
        if(typeof json === 'undefined' || json === null) {
          json = JSON.parse(JSON.stringify(fse.readJsonSync(file.path)))
        }
        var jsonValues = {}

        if(typeof request.columns !== 'undefined' && request.columns !== null && request.columns.length > 0 && request.columns[0] !== '*') {
          Array.prototype.forEach.call(request.columns, (column) => {
            if(typeof json[column] !== 'undefined' && json[column] !== null) {
              jsonValues[column] = json[column]
            }
          })
          jsonValues[config.meta.name] = json[config.meta.name]

        }else {
          jsonValues = json
        }

        if(limit < request.limit || request.limit === -1) {
          res.push(jsonValues)
        }
        limit++
      }
    })

    log.write('abe-utils-handleSqlRequest', res)
    return res
  }

  static getDataList(website, text, jsonPage) {
    var listReg = /({{abe.*type=[\'|\"]data.*}})/g,
        match

    var source = config.source.name
    if(typeof jsonPage[source] === 'undefined' || jsonPage[source] === null) {
      jsonPage[source] = {}
    }

    while (match = listReg.exec(text)) {

      var request = getAttr(match[0], 'request')
      var key = getAttr(match[0], 'key')

      // SQL LIKE REQUEST
      if(typeof request !== 'undefined' && request !== null && request !== '') {
        var data = Utils.getDataRequest(website, match[0])
        jsonPage[source][key] = data

        if(typeof jsonPage[key] !== 'undefined' && jsonPage[key] !== null) {
          var newJsonValue = []
          Array.prototype.forEach.call(jsonPage[key], (oldValue) => {
            Array.prototype.forEach.call(jsonPage[source][key], (newValue) => {
              if(oldValue[config.meta.name].link === newValue[config.meta.name].link) {
                newJsonValue.push(newValue)
              }
            })
          })
          jsonPage[key] = newJsonValue
        }
      }else if(match[0].indexOf('value=') > -1) {
        var value = Utils.getDataSource(match[0])

        if(value.indexOf('{') > -1 || value.indexOf('[') > -1) {
          try{
            value = JSON.parse(value)

            jsonPage[source][key] = value
          }catch(e){
            jsonPage[source][key] = null
            log.error('source', `${value}`, `is not a valid JSON\n${e}`)
            console.log(clc.red(`Error ${value}/is not a valid JSON`),  `\n${e}`)
          }
        }
      }
    }

    // return filesRequest
  }

  static removeDataList(text) {
    var listReg = /({{abe.*type=[\'|\"]data.*}})/g

    log.delAndWrite('Util-getList', text)
    return text.replace(listReg, '')
  }
}